/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2019 Databases and Information Systems Group TU Dortmund          *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include <iostream> // for statistics dump
#include <unordered_map>
#include <new>
#include <random>
#include <assert.h>
#include <vector>
#include <ostream>
#include "storage.h"
#include "frame.h"

namespace aidbms::storage {

    /**
     * @brief The BufferManager class implements the buffer manager, that manages
     * data pages of the database system. Template parameter buffer_size encodes the number of frames
     * (that hold pages). I.e., 256 frames allow to hold 256 pages at the same time.
     * The Policy template parameter provides the eviction strategy. An eviction policy has to provide
     * the functions
     * void onUnpin(PageID, TimeStamp),
     * void onPin(PageID, TimeStamp) and
     * Frame* evict(std::unordered_map<PageID, Frame*> &locations). See random_policy.h for an example
     * policy.
     */
    template<class Policy, std::size_t buffer_size = 256>
    class BufferManager
    {
    public:
        BufferManager() = default;
        ~BufferManager() = default;
        /**
         * @brief pin fetches a page from "storage". Calls onPin() of the selected eviction strategy.
         * @param page_id the identifier of the required page.
         * @return the pointer to the frame that holds the page.
         */
        Frame* pin(PageID page_id) {
            _eviction_policy.onPin(page_id, _pin_count++);

            // look up page in our map
            const auto location = page_locations.find(page_id);

            if (location != page_locations.end()) {
                // we already have this page in our frames
                location->second->increasePinCount();
                return location->second;
            }

            else {
                // page not loaded, try to fill empty frames
                if (page_locations.size() < buffer_size) {

                    // allocate a new frame:
                    _buffer[page_locations.size()] = Frame(page_id, _storage.load(page_id));

                    // update our map to contain the pinned page.
                    page_locations[page_id] = &_buffer[page_locations.size()];

                    return page_locations[page_id]; // return the frame pointer
                }
                else {
                    // no space; look for a victim to evict. The victim is selected
                    // accorging to the template-policy
                    Frame* victim = _eviction_policy.evict(page_locations);

                    // make sure the policy ALWAYS picks a victim
                    assert(victim);

                    // remove the victim from the map
                    page_locations.erase(victim->getCurrentPage());

                    // update statistics
                    _evict_count++;

                    // only write back when the page is marked as "dirty" (via unpin)
                    if ( victim->isDirty()) {
                        _storage.writeBack(victim->getCurrentPage(),victim->getPage());
                    }

                    // create new Frame inplace of v
                    Frame* new_page = new (victim) Frame(page_id, _storage.load(page_id));

                    // update map
                    page_locations[page_id] = new_page;

                    // return the frame pointer
                    return new_page;
                }
            }
        }
        /**
         * @brief unpin removes a page from the frames/buffer manager.
         * Note that this does not cause a page to be written back (to "storage", i.e. disk) right away.
         * This function makes an assertion that the provided id is actual present in the frame set!
         * Calls onUnpin() of the selected eviction strategy.
         * @param id identifier of the page to unpin.
         * @param was_modified mark it dirty, if this is true
         */
        void unpin(PageID id, bool was_modified) {
            const auto location = page_locations.find(id);
            assert(location != page_locations.end());

            location->second->decreasePinCount();
            location->second->becomesDirty(was_modified);

            _eviction_policy.onUnpin(id, _pin_count);
        }
        /**
         * @brief operator << prints statistic to an ostream, i.e. std::cout.
         * @param stream i.e., std::cout
         * @param buffer_manager The instance
         * @return in (modified) stream, in order to allow stream chaining.
         */
        friend std::ostream& operator<< (std::ostream& stream, const BufferManager& buffer_manager) {
            return stream
                << "Total number of page pins:      " << buffer_manager.pin_count() << "\n"
                << "Total number of page evictions: " << buffer_manager.evict_count()
            ;
        }

        /**
         * @brief pin_count gives global pin statistics.
         * @return number of calls to pin()
         */
        inline std::size_t pin_count() const {
            return _pin_count;
        }
        /**
         * @brief evict_count gives global eviction statistics.
         * @return the number of evicted pages
         */
        inline std::size_t evict_count() const {
            return _evict_count;
        }

    private:
        // All full and empty frames.
        std::array<Frame, buffer_size> _buffer;

        // Map from page_id to pointer within buffer.
        std::unordered_map<PageID, Frame*> page_locations;

        // dummy class that "symbolizes" disk access.
        Storage _storage;

        // instance of the policy, might have state!
        Policy _eviction_policy;

        std::size_t _evict_count = 0;
        std::size_t _pin_count = 0;
    };
}
