/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2019 Databases and Information Systems Group TU Dortmund          *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *------------------------------------------------------------------------------*
 */

#pragma once

#include "storage.h"

namespace aidbms::storage {

    using TimeStamp = std::size_t;
    /**
     * @brief The Frame class holds pages for the buffermanager, as well as some meta data about the page, for example
     * the dirty flag and the pin count.
     */
    class Frame {
    public:
        Frame() {
        } // only declared to allow arrow initialization

        Frame(PageID page_id, Page page) : current_page(page_id),page(page),pin_count(1),dirty(false) {
        }

        void increasePinCount() {
            pin_count += 1;
            max_pin_count = std::max(max_pin_count,pin_count);
        }

        void decreasePinCount() {
            pin_count -= 1;
        }

        bool isDirty() const {
            return dirty;
        }

        bool isUnpinned() const {
            return pin_count == 0;
        }

        void becomesDirty(bool becomes_dirty) {
            dirty |= becomes_dirty;
        }

        PageID getCurrentPage() const {
            return current_page;
        }

        Page& getPage() {
            return page;
        }

        std::size_t getMaximumPinCount() const {
            return max_pin_count;
        }

        TimeStamp getTimeOfLastAccess() const {
            return last_access;
        }

        void setTimeOfLastAccess(TimeStamp time) {
            last_access = time;
        }

    private:
        PageID current_page;
        Page page;
        std::size_t pin_count;
        bool dirty;
        std::size_t max_pin_count;
        TimeStamp last_access;
    };
}
